/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2021 (C) OpenVPMS Ltd. All Rights Reserved.
 */


package org.openvpms.component.business.dao.hibernate.im.security;

import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.hibernate.query.Query;
import org.openvpms.component.business.dao.hibernate.cache.ArchetypeIdCache;
import org.openvpms.component.business.dao.hibernate.im.common.CompoundAssembler;
import org.openvpms.component.business.dao.hibernate.im.common.Context;
import org.openvpms.component.business.dao.hibernate.im.entity.EntityLinkAssembler;
import org.openvpms.component.business.dao.hibernate.im.entity.EntityRelationshipAssembler;
import org.openvpms.component.business.dao.hibernate.im.entity.IMObjectResultCollector;
import org.openvpms.component.business.dao.hibernate.im.lookup.LookupAssembler;
import org.openvpms.component.business.dao.hibernate.im.lookup.LookupLinkAssembler;
import org.openvpms.component.business.dao.hibernate.im.lookup.LookupRelationshipAssembler;
import org.openvpms.component.business.dao.hibernate.im.party.ContactAssembler;
import org.openvpms.component.business.dao.im.security.IUserDAO;
import org.openvpms.component.business.dao.im.security.UserDAOException;
import org.openvpms.component.business.domain.im.security.User;
import org.openvpms.component.business.service.archetype.descriptor.cache.IArchetypeDescriptorCache;
import org.springframework.transaction.annotation.Transactional;

import java.util.List;


/**
 * This is an implementation of the IUserDAO DAO for hibernate. It uses the
 * Spring Framework's template classes.
 *
 * @author Jim Alateras
 */
public class UserDAOHibernate implements IUserDAO {

    /**
     * The session factory.
     */
    private final SessionFactory factory;

    /**
     * The archetype id cache.
     */
    private final ArchetypeIdCache archetypeIds;

    /**
     * The assembler.
     */
    private final CompoundAssembler assembler;

    /**
     * The archetype short names to filter on.
     */
    private String[] shortNames;

    /**
     * The query string.
     */
    private String userQuery = QUERY;

    /**
     * User query.
     */
    private static final String QUERY =
            "from " + UserDOImpl.class.getName() + " as user where user.username = :name";

    /**
     * Constructs a {@link UserDAOHibernate}.
     *
     * @param factory    the session factory
     * @param archetypes the archetype descriptor cache
     */
    public UserDAOHibernate(SessionFactory factory, IArchetypeDescriptorCache archetypes) {
        this(factory, archetypes, new ArchetypeIdCache());
    }

    /**
     * Constructs a {@link UserDAOHibernate}.
     *
     * @param factory      the session factory
     * @param archetypes   the archetype descriptor cache
     * @param archetypeIds the archetype id cache
     */
    public UserDAOHibernate(SessionFactory factory, IArchetypeDescriptorCache archetypes,
                            ArchetypeIdCache archetypeIds) {
        this.factory = factory;
        this.archetypeIds = archetypeIds;
        assembler = new Assembler(archetypes);
    }

    /**
     * Constrains users to the specified archetype short names.
     *
     * @param shortNames the user archetype short names
     */
    public void setUserArchetypes(String... shortNames) {
        this.shortNames = shortNames;
        if (shortNames != null && shortNames.length != 0) {
            StringBuilder query = new StringBuilder(QUERY);
            query.append(" and user.archetypeId.shortName in (");
            for (int i = 0; i < shortNames.length; ++i) {
                if (i > 0) {
                    query.append(',');
                }
                query.append(":shortName").append(i);
            }
            query.append(")");
            userQuery = query.toString();
        } else {
            userQuery = QUERY;
        }
    }

    /**
     * (non-Javadoc)
     *
     * @see IUserDAO#getByUserName(String)
     */
    @SuppressWarnings("unchecked")
    @Transactional(readOnly = true)
    public List<User> getByUserName(String name) {
        List<User> results;
        Session session = factory.getCurrentSession();
        try {
            Query<UserDOImpl> query = session.createQuery(userQuery, UserDOImpl.class);
            query.setParameter("name", name);
            if (shortNames != null && shortNames.length > 0) {
                for (int i = 0; i < shortNames.length; ++i) {
                    query.setParameter("shortName" + i, shortNames[i]);
                }
            }
            Context context = Context.getContext(session, assembler, archetypeIds);
            IMObjectResultCollector collector = new IMObjectResultCollector();
            collector.setContext(context);
            for (Object object : query.list()) {
                collector.collect(object);
            }
            context.resolveDeferredReferences();
            results = (List<User>) (List<?>) collector.getPage().getResults();
        } catch (Exception exception) {
            throw new UserDAOException(UserDAOException.ErrorCode.FailedToFindUserRecordByName,
                                       new Object[]{name}, exception);
        }
        return results;
    }

    private static class Assembler extends CompoundAssembler {

        Assembler(IArchetypeDescriptorCache archetypes) {
            super(archetypes);
            addAssembler(new ArchetypeAuthorityAssembler());
            addAssembler(new EntityLinkAssembler());
            addAssembler(new EntityRelationshipAssembler());
            addAssembler(new LookupAssembler());
            addAssembler(new LookupRelationshipAssembler());
            addAssembler(new LookupLinkAssembler());
            addAssembler(new SecurityRoleAssembler());
            addAssembler(new UserAssembler());
            addAssembler(new ContactAssembler());
        }
    }
}
