/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2021 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.web.component.im.delete;

import org.junit.Before;
import org.junit.Test;
import org.openvpms.archetype.rules.act.ActStatus;
import org.openvpms.archetype.rules.finance.account.FinancialTestHelper;
import org.openvpms.archetype.rules.patient.PatientArchetypes;
import org.openvpms.archetype.test.TestHelper;
import org.openvpms.archetype.test.builder.doc.TestDocumentFactory;
import org.openvpms.component.business.domain.im.act.Act;
import org.openvpms.component.business.domain.im.act.FinancialAct;
import org.openvpms.component.business.domain.im.common.Entity;
import org.openvpms.component.business.domain.im.common.IMObject;
import org.openvpms.component.business.domain.im.datatypes.quantity.Money;
import org.openvpms.component.business.domain.im.party.Party;
import org.openvpms.component.business.domain.im.product.Product;
import org.openvpms.component.business.service.archetype.helper.ActBean;
import org.openvpms.web.component.app.Context;
import org.openvpms.web.component.app.LocalContext;
import org.openvpms.web.component.im.edit.IMObjectEditor;
import org.openvpms.web.echo.help.HelpContext;
import org.openvpms.web.test.AbstractAppTest;
import org.springframework.beans.factory.annotation.Autowired;

import java.math.BigDecimal;
import java.util.List;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;


/**
 * Tests the {@link ConfirmingDeleter} class.
 *
 * @author Tim Anderson
 */
public class ConfirmingDeleterTestCase extends AbstractAppTest {

    /**
     * The help context.
     */
    private final HelpContext help = new HelpContext("foo", null);

    /**
     * The context.
     */
    private final Context context = new LocalContext();

    /**
     * The document factory.
     */
    @Autowired
    private TestDocumentFactory documentFactory;

    /**
     * The deletion handler factory.
     */
    private IMObjectDeletionHandlerFactory factory;


    /**
     * Sets up the test case.
     */
    @Before
    public void setUp() {
        super.setUp();
        factory = new IMObjectDeletionHandlerFactory(getArchetypeService());
        factory.setApplicationContext(applicationContext);
    }

    /**
     * Verifies that attempting to delete an entity with participations deactivates it.
     */
    @Test
    public void testDeleteEntityWithParticipations() {
        // create a customer and associated invoice
        Party customer = TestHelper.createCustomer();
        Party pet = TestHelper.createPatient(customer);
        List<FinancialAct> invoice = FinancialTestHelper.createChargesInvoice(new Money("100"),
                                                                              customer, pet,
                                                                              TestHelper.createProduct(),
                                                                              ActStatus.POSTED);
        save(invoice);     // customer has participation relationships to the invoice

        TestDeleter<IMObject> deleter = new TestDeleter<>();
        TestListener<IMObject> listener = new TestListener<>();
        deleter.delete(customer, context, help, listener);

        // verify the customer has been deactivated rather than deleted
        checkDeleter(deleter, false, true, false, false);
        checkListener(listener, false);

        customer = get(customer);
        assertNotNull(customer);
        assertFalse(customer.isActive());

        // now attempt deletion again. The deactivated() method should be invoked
        deleter = new TestDeleter<>();
        listener = new TestListener<>();
        deleter.delete(customer, context, help, listener);
        checkDeleter(deleter, false, false, true, false);
        checkListener(listener, false);
    }

    /**
     * Verifies that attempting to delete an entity which is the source of a relationship deactivates it instead.
     */
    @Test
    public void testDeleteSourceWithEntityRelationships() {
        Party customer = TestHelper.createCustomer();
        Party pet = TestHelper.createPatient(customer);

        TestDeleter<IMObject> deleter = new TestDeleter<>();
        TestListener<IMObject> listener = new TestListener<>();
        deleter.delete(customer, context, help, listener);

        // verify the customer has been deactivated
        checkDeleter(deleter, false, true, false, false);
        checkListener(listener, false);

        customer = get(customer);
        assertNotNull(customer);
        assertFalse(customer.isActive());
        assertNotNull(get(pet));
    }

    /**
     * Verifies that attempting to delete an entity which is the target of a relationship
     * invokes {@link AbstractIMObjectDeleter#delete}, and performs the removal.
     */
    @Test
    public void testDeleteTargetWithEntityRelationships() {
        Party customer = TestHelper.createCustomer();
        Party pet = TestHelper.createPatient(customer);

        TestDeleter<IMObject> deleter = new TestDeleter<>();
        TestListener<IMObject> listener = new TestListener<>();
        deleter.delete(pet, context, help, listener);

        // verify the customer has been deleted
        checkDeleter(deleter, true, false, false, false);
        checkListener(listener, true);

        assertNull(get(pet));
        assertNotNull(get(customer));
    }

    /**
     * Verifies that <em>entity.documentTemplate</em> can be deleted despite having an
     * <em>participation.document</em> participation.
     */
    @Test
    public void testDeleteTemplate() {
        TestDeleter<IMObject> deleter = new TestDeleter<>();

        // create a template with associated act.documentTemplate
        Entity template = (Entity) documentFactory.createTemplate(PatientArchetypes.DOCUMENT_FORM);
        TestListener<IMObject> listener = new TestListener<>();
        deleter.delete(template, context, help, listener);

        // verify the template has been deleted
        checkDeleter(deleter, true, false, false, false);
        checkListener(listener, true);

        assertNull(get(template));
    }

    /**
     * Verifies that <em>entity.documentTemplate</em> can't be deleted if it has participations
     * other than <em>participation.document</em>. These will be deactivated.
     */
    @Test
    public void testDeleteTemplateWithParticipations() {
        TestDeleter<IMObject> deleter = new TestDeleter<>();

        Entity template = (Entity) documentFactory.createTemplate(PatientArchetypes.DOCUMENT_FORM);
        Act act = (Act) create(PatientArchetypes.DOCUMENT_FORM);
        Party patient = TestHelper.createPatient();
        ActBean bean = new ActBean(act);
        bean.addNodeParticipation("patient", patient);
        bean.addNodeParticipation("documentTemplate", template);
        save(act);

        TestListener<IMObject> listener = new TestListener<>();
        deleter.delete(template, context, help, listener);

        // verify the template has been deactivated
        checkDeleter(deleter, false, true, false, false);
        checkListener(listener, false);

        template = get(template);
        assertNotNull(template);
        assertFalse(template.isActive());

        // now attempt deletion again. The deactivated() method should be invoked
        deleter = new TestDeleter<>();
        listener = new TestListener<>();
        deleter.delete(template, context, help, listener);
        checkDeleter(deleter, false, false, true, false);
        checkListener(listener, false);
    }

    /**
     * Verifies that attempting to delete an entity that is the target of an entity link deactivates it.
     */
    @Test
    public void testDeleteWithEntityLinks() {
        Party location = TestHelper.createLocation();
        TestHelper.createCustomer(location);

        TestDeleter<IMObject> deleter = new TestDeleter<>();
        TestListener<IMObject> listener = new TestListener<>();

        deleter.delete(location, context, help, listener);

        checkDeleter(deleter, false, true, false, false);
        checkListener(listener, false);

        location = get(location);
        assertNotNull(location);
        assertFalse(location.isActive());
    }

    /**
     * Verifies POSTED acts can't be deleted or deactivated.
     */
    @Test
    public void testDeletePostedAct() {
        Party customer = TestHelper.createCustomer();
        Product product = TestHelper.createProduct();
        List<FinancialAct> charge = FinancialTestHelper.createChargesCounter(BigDecimal.TEN, customer, product,
                                                                             ActStatus.POSTED);
        save(charge);

        TestDeleter<IMObject> deleter = new TestDeleter<>();
        TestListener<IMObject> listener = new TestListener<>();
        deleter.delete(charge.get(0), context, help, listener);

        checkDeleter(deleter, false, false, false, true);
        checkListener(listener, false);

        FinancialAct act = get(charge.get(0));
        assertNotNull(act);
        assertTrue(act.isActive());
    }

    /**
     * Verifies that the appropriate deleter methods have been invoked.
     *
     * @param deleter     the deleter to check
     * @param remove      if {@code true} expect remove() to have been invoked
     * @param deactivate  if {@code true} expect deactivate() to have been invoked
     * @param deactivated if {@code true} expect deactivated() to have been invoked
     * @param unsupported if {@code true} expect unsupported() to have been invoked
     */
    private void checkDeleter(TestDeleter<IMObject> deleter, boolean remove, boolean deactivate, boolean deactivated,
                              boolean unsupported) {
        assertEquals(remove, deleter.removeInvoked());
        assertEquals(deactivate, deleter.deactivateInvoked());
        assertEquals(deactivated, deleter.deactivatedInvoked());
        assertEquals(unsupported, deleter.unsupportedInvoked());
    }

    /**
     * Verifies that the appropriate listener methods have been invoked.
     *
     * @param listener the listener to check
     * @param deleted  if {@code true} expect the deleted() method to have been invoked
     */
    private void checkListener(TestListener listener, boolean deleted) {
        assertEquals(deleted, listener.deletedInvoked());
        assertFalse(listener.failedInvoked());
    }

    private class TestDeleter<T extends IMObject> extends ConfirmingDeleter<T> {

        /**
         * Determines if remove() was invoked.
         */
        private boolean remove;

        /**
         * Determines if deactivate() has been invoked.
         */
        private boolean deactivate;

        /**
         * Determines if deactivated() has been invoked.
         */
        private boolean deactivated;

        /**
         * Determines if unsupported() has been invoked.
         */
        private boolean unsupported;

        /**
         * Constructs a {@code TestDeleter}.
         */
        public TestDeleter() {
            super(factory);
        }

        /**
         * Determines if remove() has been invoked.
         *
         * @return {@code true} if remove() was invoked
         */
        public boolean removeInvoked() {
            return remove;
        }

        /**
         * Determines if deactivate() has been invoked.
         *
         * @return {@code true} if deactivate() was invoked
         */
        public boolean deactivateInvoked() {
            return deactivate;
        }

        /**
         * Determines if deactivated() has been invoked.
         *
         * @return {@code true} if deactivated() was invoked
         */
        public boolean deactivatedInvoked() {
            return deactivated;
        }

        /**
         * Determines if unsupported() has been invoked.
         *
         * @return {@code true} if unsupported() was invoked
         */
        public boolean unsupportedInvoked() {
            return unsupported;
        }

        /**
         * Invoked to delete an object.
         *
         * @param handler  the deletion handler
         * @param context  the context
         * @param help     the help context
         * @param listener the listener to notify
         */
        @Override
        protected void delete(IMObjectDeletionHandler<T> handler, Context context, HelpContext help,
                              IMObjectDeletionListener<T> listener) {
            remove = true;
            doDelete(handler, context, help, listener);
        }

        /**
         * Invoked to deactivate an object.
         *
         * @param handler  the deletion handler
         * @param listener the listener
         * @param help     the help context
         */
        @Override
        protected void deactivate(IMObjectDeletionHandler<T> handler, IMObjectDeletionListener<T> listener,
                                  HelpContext help) {
            deactivate = true;
            doDeactivate(handler, listener);
        }

        /**
         * Invoked when an object cannot be de deleted, and has already been deactivated.
         *
         * @param object the object
         * @param help   the help context
         */
        @Override
        protected void deactivated(T object, HelpContext help) {
            deactivated = true;
        }

        /**
         * Invoked when deletion and deactivation of an object is not supported.
         *
         * @param object   the object
         * @param reason   reason the object couldn't be deleted. May be {@code null}
         * @param listener the listener
         */
        @Override
        protected void unsupported(T object, String reason, IMObjectDeletionListener<T> listener) {
            unsupported = true;
        }
    }

    private class TestListener<T extends IMObject> extends AbstractIMObjectDeletionListener<T> {

        /**
         * Determines if deleted() was invoked.
         */
        private boolean deleted;

        /**
         * Determines if one of the failed*() methods was invoked.
         */
        private boolean failed;

        /**
         * Determines if deleted() was invoked.
         *
         * @return {@code true} if deleted() was invoked
         */
        public boolean deletedInvoked() {
            return deleted;
        }

        /**
         * Determines if one of the failed*() methods was invoked.
         *
         * @return {@code true} if a failed*() method was invoked
         */
        public boolean failedInvoked() {
            return failed;
        }

        @Override
        public void deleted(T object) {
            deleted = true;
        }

        @Override
        public void failed(T object, Throwable cause) {
            failed = true;
        }

        @Override
        public void failed(T object, Throwable cause, IMObjectEditor editor) {
            failed = true;
        }
    }
}
